/**
 * Ext GWT 3.0.0-rc2 - Ext for GWT
 * Copyright(c) 2007-2012, Sencha, Inc.
 * licensing@sencha.com
 *
 * http://sencha.com/license
 */
package com.sencha.gxt.desktopapp.client.spreadsheet;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import com.google.gwt.core.client.GWT;
import com.google.gwt.safehtml.shared.SafeHtmlUtils;
import com.google.web.bindery.autobean.shared.AutoBean;
import com.google.web.bindery.autobean.shared.AutoBeanCodex;
import com.google.web.bindery.autobean.shared.AutoBeanUtils;
import com.sencha.gxt.data.shared.ListStore;
import com.sencha.gxt.data.shared.ModelKeyProvider;
import com.sencha.gxt.widget.core.client.grid.ColumnConfig;
import com.sencha.gxt.widget.core.client.grid.RowNumberer;

public class Worksheet implements TableValueProvider {

  private static final int DEFAULT_ROW_COUNT = 2;
  private static final int DEFAULT_COLUMN_COUNT = 2;

  private TableFactory tableFactory;
  private ListStore<Row> listStore;
  private ColumnProvider columnProvider;
  private ModelKeyProvider<Row> modelKeyProvider;
  private int nextRowId;

  public Worksheet(ColumnProvider columnProvider) {
    this.columnProvider = columnProvider;
  }

  public void changeColumnCount(int newColumnCount) {
    List<Row> rows = getListStore().getAll();
    int rowCount = rows.size();
    for (int rowIndex = 0; rowIndex < rowCount; rowIndex++) {
      List<String> columns = rows.get(rowIndex).getColumns();
      if (newColumnCount < columns.size()) {
        rows.get(rowIndex).setColumns(new ArrayList<String>(columns.subList(0, newColumnCount)));
      } else {
        createColumns(columns, newColumnCount);
      }
    }
  }

  public void changeRowCount(int newRowCount) {
    if (newRowCount < getRowCount()) {
      List<Row> rows = getListStore().getAll();
      getListStore().replaceAll(new ArrayList<Row>(rows.subList(0, newRowCount)));
    } else {
      createRows(newRowCount, getColumnCount());
    }
  }

  public double evaluate(String expression) {
    Evaluator evaluator = new Evaluator(expression.substring(1), this);
    return evaluator.evaluateExpression();
  }

  public double evaluate(String expression, int rowIndex, int columnIndex) {
    return evaluateAbsolute(expression, rowIndex, columnIndex - 1);
  }

  public int getColumnCount() {
    return getRowCount() == 0 ? 0 : getListStore().get(0).getColumns().size();
  }

  public ListStore<Row> getListStore() {
    if (listStore == null) {
      listStore = new ListStore<Row>(getModelKeyProvider());
      listStore.setAutoCommit(true);
    }
    return listStore;
  }

  public String getNameAbsolute(int rowIndex, int columnIndex) {
    String name = null;
    ColumnConfig<Row, Object> columnConfig = getColumnProvider().getColumn(columnIndex);
    if (!(columnConfig instanceof RowNumberer)) {
      name = columnConfig.getHeader().asString() + (rowIndex + 1);
    }
    return name;
  }

  public int getRowCount() {
    return getListStore().size();
  }

  public String getValue() {
    nextRowId = 0;
    List<Row> rows = new LinkedList<Row>();

    int rowCount = getRowCount();
    int columnCount = getColumnCount();

    for (int rowIndex = 0; rowIndex < rowCount; rowIndex++) {
      List<String> columns = new LinkedList<String>();
      for (int columnIndex = 0; columnIndex < columnCount; columnIndex++) {
        columns.add(getValue(rowIndex, columnIndex));
      }
      Row row = getTableFactory().row().as();
      row.setId(nextRowId++);
      row.setColumns(columns);
      rows.add(row);
    }

    Table table = getTableFactory().table().as();
    table.setNextRowId(nextRowId);
    table.setRows(rows);

    AutoBean<Table> tableAutoBean = AutoBeanUtils.getAutoBean(table);
    String value = AutoBeanCodex.encode(tableAutoBean).getPayload();
    return value;
  }

  public String getValue(int rowIndex, int columnIndex) {
    return getValueAbsolute(rowIndex, columnIndex + 1);
  }

  public String getValueAbsolute(int rowIndex, int columnIndex) {
    Row row = getListStore().get(rowIndex);
    ColumnConfig<Row, Object> columnConfig = getColumnProvider().getColumn(columnIndex);
    return columnConfig.getValueProvider().getValue(row).toString();
  }

  public void renameColumns() {
    int columnCount = getColumnCount();
    for (int columnIndex = 0; columnIndex < columnCount; columnIndex++) {
      String columnName = SpreadsheetUtilities.getColumnName(columnIndex);
      getColumnProvider().setColumnHeader(columnIndex + 1, SafeHtmlUtils.fromString(columnName));
    }
  }

  public void setValue(String value) {
    listStore = null;
    if (value == null || value.isEmpty()) {
      nextRowId = 0;
      createRows(DEFAULT_ROW_COUNT, DEFAULT_COLUMN_COUNT);
    } else {
      Table table = AutoBeanCodex.decode(getTableFactory(), Table.class, value).as();
      nextRowId = table.getNextRowId();
      List<Row> rows = table.getRows();
      getListStore().addAll(rows);
    }
  }

  public void setValueAbsolute(int rowIndex, int columnIndex, String value) {
    Row row = getListStore().get(rowIndex);
    ColumnConfig<Row, Object> columnConfig = getColumnProvider().getColumn(columnIndex);
    columnConfig.getValueProvider().setValue(row, value);
  }

  protected double evaluateAbsolute(String expression, int rowIndex, int columnIndex) {
    Evaluator evaluator = new Evaluator(expression.substring(1), this);
    evaluator.setRowIndex(rowIndex);
    evaluator.setColumnIndex(columnIndex);
    return evaluator.evaluateExpression();
  }

  private void createColumns(List<String> columns, int newColumnCount) {
    while (columns.size() < newColumnCount) {
      columns.add("");
    }
  }

  private void createRows(int newRowCount, int newColumnCount) {
    while (getRowCount() < newRowCount) {
      ArrayList<String> columns = new ArrayList<String>();
      createColumns(columns, newColumnCount);
      Row row = getTableFactory().row().as();
      row.setColumns(columns);
      row.setId(getNextRowId());
      getListStore().add(row);
    }
  }

  private ColumnProvider getColumnProvider() {
    return columnProvider;
  }

  private ModelKeyProvider<Row> getModelKeyProvider() {
    if (modelKeyProvider == null) {
      modelKeyProvider = new ModelKeyProvider<Row>() {
        @Override
        public String getKey(Row row) {
          return Integer.toString(row.getId());
        }
      };
    }
    return modelKeyProvider;
  }

  private int getNextRowId() {
    return nextRowId++;
  }

  private TableFactory getTableFactory() {
    if (tableFactory == null) {
      tableFactory = GWT.create(TableFactory.class);
    }
    return tableFactory;
  }

}
